;===================================================================
;
;                          resstrat.scm
;
;                   Gandalf version c-2.1
;
;
;    
;               Special stragies, strategy selection 
;  
;  Other files: resbase.scm
;              (basic operations, incl subst. unif. and subsumption)
;               resmain.scm
;               resio.scm
;
;====================================================================



;=====================================================================

; Hyper-contradiction tries to produce an immediate contradiction from cl
; using (generalised) hyperresolution.
; If that fails, nothing is produced.
;
; cl is always longer than one literal (otherwise a contradiction
; should have been found earlier). If cl contains ground literals,
; it is not checked (ground literals should have been uniit-deleted).
;
; called in handle-derived-inf-cl-pair-aux
; iff *hyper-contradiction-strategy* is #t

(define (hyper-contradiction? fullcl)
  (if (and (not *foundflag*)
           (not (empty-cl? (cdr fullcl))) 
	   (not (unit-cl? (cdr fullcl)))
	   (every? (lambda(lit)(term-may-cont-vars? lit)) (cdr fullcl)) )
      (begin
	(if (>= *display-info-level* 70)
	    (begin
	      (newline)
	      (display "hyper-contradiction? entered: ")
	      (print-full-clause fullcl)
	      (newline) ))
	(set!  *unit-clash-flag* #t) ; ONLY #t during hyper-contradiction
	(let* ((ycl (map rename-x-to-y-term (cdr fullcl)))
	       (yinf (car fullcl))
	       (ylitinf (caddr yinf)) 
	       (one-step-results (one-step-for-hyper (cons yinf ycl))) )
	  ; loop for all partial (one-step) results
	  (do ((results-part one-step-results (cdr results-part)))
	      ((or (null? results-part) 
		   *foundflag*
		   *stop-fast-loop-flag*
	           (and (if  (and (not *stop-fast-loop-flag*)
				  (> (- (get-check-time 63) *slow-start-time*)
				     *one-pass-time*))
			     (begin
			       (store-time-checkpoint! 63)
			       (set! *stop-fast-loop-flag* #t)
			       #t)))))
	    (hyper-contradiction? (car results-part)) ))
	(set! *unit-clash-flag* #f) ; ONLY #t during hyper-contradiction
	*foundflag*)
      #f))

; full-hyper-contradiction? allows fullcl to contain also ground
; literals 


(define (full-hyper-contradiction? fullcl)
    (if (and (not *foundflag*)
           (not (empty-cl? (cdr fullcl))) 
	   (not (unit-cl? (cdr fullcl))) )
      (begin
	(if (>= *display-info-level* 70)
	    (begin
	      (newline)
	      (display "full-hyper-contradiction? entered: ")
	      (print-full-clause fullcl)
	      (newline) ))
	(set!  *unit-clash-flag* #t) ; ONLY #t during hyper-contradiction
	(let* ((ycl (map rename-x-to-y-term (cdr fullcl)))
	       (yinf (car fullcl))
	       (ylitinf (caddr yinf)) 
	       (one-step-results (one-step-for-hyper (cons yinf ycl))) )
	  ; loop for all partial (one-step) results
	  (do ((results-part one-step-results (cdr results-part)))
	      ((or (null? results-part) 
		   *foundflag*
		   *stop-fast-loop-flag*
		   (and (if  (and (not *stop-fast-loop-flag*)
				  (> (- (get-check-time 62) *slow-start-time*)
				     *one-pass-time*))
			     (begin
			       (store-time-checkpoint! 62)
			       (set! *stop-fast-loop-flag* #t)
			       #t)))))
	    (hyper-contradiction? (car results-part)) ))
	(set! *unit-clash-flag* #f) ; ONLY #t during hyper-contradiction
	*foundflag*)
      #f))

; start-binary-hypercheck-strategy first checks all the existing
; clauses for immediate hyper-contradiction and then allows the
; ordinary binary resolution to continue with full immediate check
; of each derived clause


(define (start-binary-hypercheck-strategy)
  (let ()
  ; first for usable
    (do ((part *usable-clauses-list* (cdr part)))
	((or (null? part) 
	     *foundflag*
	     *stop-fast-loop-flag*))
      (if (and (pair? (car part))
	       (not (unit-cl? (car part))) )
	  (full-hyper-contradiction? (car part)) ))
    ; second for sos
    (do ((part *sos-clauses-list*  (cdr part)))
	((or (null? part) 
	     *foundflag*
	     *stop-fast-loop-flag*))
      (if (and (pair? (car part))
	       (not (unit-cl? (car part))) )
	  (full-hyper-contradiction? (car part))  ))
    ; set the full-check strategy 
    (set! *hyper-contradiction-strategy* #t) ))
	


; start-backward-hyper-analysis retains only fully negative
; clauses in usable, all others are moved to sos. 
; Only nonunit clauses are then retained in sos (unit clauses
; removed - they are left only in *unit-alist*).
; All non-answer literals are made resolvable (no ordering!)

(define (start-backward-hyper-analysis)

;(newline)
;(display "units:")
;(newline)
;(for-each (lambda(y)
;	    (newline)
;	    (for-each 
;	     (lambda(x) (if (pair? x) (print-clause x))(newline))
;	     (cdr y) ))
;	  *unit-alist*)


;(newline)
;(display "usable:")
;(newline)
; (for-each (lambda(x)(if (pair? x) (print-clause x))(newline))  
;*usable-clauses-list*)

;(newline)
;(display "sos:")
;(newline)
;(for-each (lambda(x)(if (pair? x) (print-clause x))(newline))  
;*sos-clauses-list*)

  (let ((newsos '())
	(newusable '()) 
	(cl '()) )
    
    ; set strategic flags
    (init-binres-strat)  
    (set! *hyper-contradiction-strategy* #t)

    (set! *unit-strategy-flag* #f)
    ; (set! *factors-stratflag* #f)
    (set! *factors-stratflag* #t)
    (set! *ordered-factor-flag* #f)
    (set! *order-strategy-flag* #f) 
    (set! *poslitpref-order-strategy-flag* #t)
    (set! *name-order-strategy-flag* #f) 
    ; (set! *paramodulation-allowed-flag* #f)
    (set! *paramodulation-allowed-flag* #t)
    (set! *remove-identical-lit-stratflag* #t)

    (set! *sos-preference-cur-ratio* 2)

    ; clear nonunit subsumption stuff

    (set! *nonunit-discr-tree* (list '())) 

;(trace havealready-clause? )
;(trace check-nonunit-discr-tree)
;(trace   one-step-for-hyper)
;(trace hyper-contradiction?)
;(trace given-clause-slow-loop)
;(trace usable-clauses-fast-loop)
;(trace two-clause-binres-loop)
;(trace one-step-for-hyper)
;(trace resolve-unify-ordinary)
;(trace make-poslit-preferred!)
;(trace mixed-cl?)
;(trace make-all-resolvable!)
;(trace handle-derived-inf-cl-pair-aux )
;(trace handle-derived-inf-cl-pair)
;(trace insert-clause-discr-tree!) 

;(set! *display-info-level* 200)
;(trace full-unit-cutoff!)
;(trace full-unit-cutoff-double! )

    ; filter usable
    (do ((part *usable-clauses-list* (cdr part)))
	((or *foundflag*
	     (null? part) ))
      (set! cl (car part)) 
      (if (pair? cl)(make-poslit-preferred! cl)) 
      (cond ((not (pair? cl)))
	    ((negative-cl? (cdr cl))
  	         (hyper-contradiction? cl) ; test for immediate contradiction!
		 (set! newsos (cons cl newsos)) )
	     ; ((unit-cl? (cdr cl))) ; remove nonnegative units
	     ; ((unit-cl? (cdr cl))) ; remove nonnegative units
	     (else 
	        (set! newusable (cons cl newusable)) )))
    (set! *usable-clauses-list* newusable)
  

    ; filter sos
    
    (do ((part *sos-clauses-list* (cdr part)))
	((or *foundflag*
	     (null? part)))
      (set! cl (car part))
      (if (pair? cl)(make-poslit-preferred! cl)) 
      (cond 
        ((not (pair? cl)))
	((negative-cl? (cdr cl))
	    (hyper-contradiction? cl)
            (set! newsos (cons cl newsos)) )
	(else
	    (set! *usable-clauses-list*
		  (cons cl *usable-clauses-list*) ))))

    (set! *sos-clauses-list* newsos)

    (if (not *int-logic-flag*)
	(set! *sos-clauses-list-end* (my-last-pair *sos-clauses-list*)) )
    ; (set! *usable-clauses-list* (cons '() *usable-clauses-list*))
    (set! *usable-clauses-list-end* (my-last-pair *usable-clauses-list*))


;(newline)
;(display "usable:")
;(newline)
; (for-each (lambda(x)(if (pair? x) (print-clause x))(newline))  
;  *usable-clauses-list*)

;(newline)
;(display "sos:")
;(newline)
; (for-each (lambda(x)(if (pair? x) (print-clause x))(newline))  
; *sos-clauses-list*)

    ; start search all over again

    ))

; make-all-resolvable! changes litinfolst of a clause destructively,
; making all non-answer-literals resolvable.

(define (make-all-resolvable! fullcl)
  (let ((tmp 0)
	(x 0) )
    (do ((lit-part (caddr (car fullcl)) (cdr lit-part)))
	((null? lit-part))
      (set! x (car lit-part))
      (if (zero? (logand *litinf-isanswer-mask* (car x)))
	  (begin
	    (set! tmp (logxor *litinf-nouse-mask* 
			  (logior *litinf-nouse-mask* (car x)) ))
	    (set-car! x tmp) )))))

; make-poslit-preferred! changes litinfolst of a clause destructively,
; making positive literals preferred.


(define (make-poslit-preferred! fullcl)
  (if (not (mixed-cl? (cdr fullcl)))
      ; non-mixed clauses are fully cuttable
      (make-all-resolvable! fullcl)
      ; we have a mixed clause
      (do ((info-part (caddr (car fullcl)) (cdr info-part))
	   (lit-part (cdr fullcl) (cdr lit-part)) )
	  ((null? info-part))
	  (if (zero? (logand *litinf-isanswer-mask* (caar info-part)))
	      (if (neg-fun? (caar lit-part))
		  ; negative literals prohibited
		  (set-car! 
		   (car info-part)
		   (logior *litinf-nouse-mask* (caar info-part)) )
		  ; positive literals allowed
		  (set-car!
		   (car info-part)
		   (logxor *litinf-nouse-mask* 
			   (logior *litinf-nouse-mask* 
				   (caar info-part) ))))))))



; =========================  END ==================================